!*Copyright 2002 Stanford University
!*All rights reserved
!
!*Author: Antoine Guitton
program Createmis

  use createhelixmod
  use helix
  use bound
  use sep 

  implicit none

  integer                         :: n(5)  
  real                            :: o(5),d(5) 

  !SIGNAL FILTER PARAMETERS
  integer                         :: na_signal(3),center_signal(3),gap_signal(3) !filter size
  integer                         :: npatch_signal(3)                            !number of patches
  integer                         :: ncoef_signal

  !NOISE FILTER PARAMETERS
  integer                         :: na_noise(3),center_noise(3),gap_noise(3) !filter size
  integer                         :: npatch_noise(3)                          !number of patches
  integer                         :: ncoef_noise

  type (filter)                   :: Npef, Spef
  real, dimension(:), allocatable :: rr

  call initpar()

  !read parameters
  call read_general_pars_C(n,o,d)

  !signal filter
  call read_filter_pars_C("noise",na_noise,center_noise,gap_noise,npatch_noise)
  call read_filter_pars_C("signal",na_signal,center_signal,gap_signal,npatch_signal)

  Npef =  createhelix( n(1:3), center_noise, gap_noise,  na_noise)
  Spef =  createhelix( n(1:3), center_signal,gap_signal, na_signal)

  call boundn   ( n(1:3)  , n(1:3), na_noise,  Npef)
  call boundn   ( n(1:3)  , n(1:3), na_signal, Spef)

  allocate(rr(product(n)))

  rr=1.
  where(Npef%mis) rr=0.

  call srite("out",rr,size(rr)*4)
  rr=1.

  where(Spef%mis) rr=0.
  
  call srite("out",rr,size(rr)*4)

  call to_history('n4',2) 

  deallocate(rr)

end program Createmis

subroutine read_general_pars_C(n,o,d)
  
  use sep
 
  implicit none
 
  integer             :: n(5)
  real                :: o(5), d(5)
  integer             :: ndim, i
  character(len=1024) :: label
  integer,external    :: sep_get_number_data_axes,auxin

  !first DEAL with input data
  
  n(3)=1
  n(4)=1
  n(5)=1

  if (0.ne.sep_get_number_data_axes("in",ndim)) then
     call seperr("trouble obtaining data axes \n")
  end if

  if (ndim<2 .or. ndim>5) then
     call seperr("ndim must be between 2 and 5\n")
  end if
  
  do i=1,ndim
     call sep_get_data_axis_par("in",i,n(i),o(i),d(i),label)
  end do
    
end subroutine read_general_pars_C

subroutine read_filter_pars_C(string,na,center,gap,npatch)
  
  use sep
  
  implicit none
  
  character            (len=*) string
  integer              :: na(3),center(3),gap(3),npatch(3)
  character (len=1024) ::  temp
  integer,external     :: getch
  
  temp="na_"//string
  call from_param(temp,na,(/5,3,3/))
  
  temp="gap_"//string
  call from_param(temp,gap,(/0,0,0/))
  
  temp="center_"//string
  call from_param(temp,center,(/floor(real(na(1))/2.+.5),floor(real(na(2))/2+.5),1/))

  temp="npatch_"//string
  call from_param(temp,npatch,(/10,10,1/))

end subroutine read_filter_pars_C
