! Author: Antoine Guitton 06 August 2002
!
! Creates the NS filters and rougheners
! according to the number of patches.
! Works for 2D NS filters only

module build_filter2d_mod
  
  use createnhelixmod
  use lapfac
  
  implicit none
  
  integer,private,save   :: n(2)            !data size
  integer,private,save   :: npatch(2)       !number of patches
  integer,private,save   :: center(2),na(2) !box size and zero lag
  real,   private,save   :: o(2),d(2)       ! sampling of data
  
contains

  subroutine init_build_filters2d(n_in,center_in,na_in, &
       & o_in,d_in,npatch_in)
    
    integer :: n_in(2),npatch_in(2),center_in(2),na_in(2)
    real    :: o_in(2),d_in(2)
    
    n=n_in;npatch=npatch_in;center=center_in
    na=na_in;o=o_in;d=d_in
    
  end subroutine init_build_filters2d
  
  subroutine build_filters2d(filt,rough,ncoef)
    
    type(nfilter)        :: filt,rough  ! filter and rough
    integer              :: ncoef
   
    call build_cartesian_2d_space(filt,npatch,ncoef)
    call create_lap_2d_rough    (rough,npatch,ncoef)
    
  end subroutine build_filters2d

  subroutine build_filters2d_smp(filt,ncoef)
    
    type(nfilter)        :: filt ! filter
    integer              :: ncoef
   
    call build_cartesian_2d_space(filt,npatch,ncoef)
    
  end subroutine build_filters2d_smp

  subroutine build_cartesian_2d_space(filt,npatch,ncoef)
    
    type(nfilter)        :: filt
    integer              :: npatch(2),ncoef
    integer              :: i1,i,i2,iref,iout(2),j(2)
    integer, allocatable :: map(:)
   
    allocate(map(product(n)))

    j(1) = ceiling(real(n(1))/real(npatch(1))) ! number of coefficients
    j(2) = ceiling(real(n(2))/real(npatch(2))) ! per filter in the 2 directions
    
    do i2=1,n(2)
       iout(2)=ceiling(real(i2)/real(j(2)))
       do i1=1,n(1)
          iout(1)=ceiling(real(i1)/real(j(1)))
          i=i1+n(1)*(i2-1)
          iref=iout(1)+npatch(1)*(iout(2)-1)
          map(i)=iref                          ! map(i) is the table for the patches
       end do
    end do
    filt = createnhelix(n,center,(/0,0/),na,map)
    ncoef= size(filt%hlx(1)%lag)
    deallocate(map)
    
  end subroutine build_cartesian_2d_space
  
  subroutine create_lap_2d_rough(rough,npatch,ncoef)
    
    type(nfilter)        :: rough
    integer              :: ncoef,npatch(2)
    integer              :: npts(2)
    integer, allocatable :: nh(:),map(:)
    
    npts=(/npatch(1),npatch(2)/)
    allocate(nh(product(npts)))
    allocate(map(ncoef*product(npts)))
    
    nh=ncoef
    map=1
    
    nh=min(6,npts(1))*2
    call nallocate(rough,nh,map)
    rough%hlx(1)=lapfac2(.01,npts(1),nh(1)/2)
    
    deallocate(nh,map)
    
  end subroutine create_lap_2d_rough
  
end module build_filter2d_mod














