! Author: Antoine Guitton 06 August 2002
!
! Creates the NS filters and rougheners
! according to the number of patches.
! Works for 3D NS filters only

module build_filter3d
  
  use createnhelixmod_smp
  use createnhelixmod
  use lapfac3
  
  implicit none
  
  integer,private,save   :: n(3)            !data size
  integer,private,save   :: npatch(3)       !number of patches
  integer,private,save   :: center(3),na(3) !box size and zero lag
  real,   private,save   :: o(3),d(3)       ! sampling of data
  real,   dimension(:), pointer , private :: nmis
contains

  subroutine init_build_filters3d(n_in,center_in,na_in, &
       & o_in,d_in,npatch_in)
    
    integer :: n_in(3),npatch_in(3),center_in(3),na_in(3)
    real    :: o_in(3),d_in(3)
    
    n=n_in;npatch=npatch_in;center=center_in
    na=na_in;o=o_in;d=d_in
    
  end subroutine init_build_filters3d
  
  subroutine init_build_filters3d_smp(n_in,center_in,na_in, &
  & o_in,d_in,npatch_in,nmis_in)
    
    integer :: n_in(3),npatch_in(3),center_in(3),na_in(3)
    real    :: o_in(3),d_in(3)
    real, dimension(:), pointer :: nmis_in
 
    n=n_in;npatch=npatch_in;center=center_in
    na=na_in;o=o_in;d=d_in
    
    allocate(nmis(product(n))); nmis=nmis_in 

  end subroutine init_build_filters3d_smp
  
  subroutine build_filters3d(filt,rough,ncoef)
    
    type(nfilter)        :: filt,rough  ! filter and rough
    integer              :: ncoef
    
    call build_cartesian_3d_space(filt,npatch,ncoef)
    call create_lap_3d_rough    (rough,npatch,ncoef)
    
  end subroutine build_filters3d

  subroutine build_filters3d2(filt,rough,ncoef)
    
    type(nfilter)        :: filt,rough  ! filter and rough
    integer              :: ncoef

    call build_cartesian_3d_space_smp(filt,npatch,ncoef)
    call create_lap_3d_rough    (rough,npatch,ncoef)
    
  end subroutine build_filters3d2

  subroutine build_filters3d_smp(filt,ncoef)
    
    type(nfilter)        :: filt   ! filter
    integer              :: ncoef
    
    call build_cartesian_3d_space(filt,npatch,ncoef)
    
  end subroutine build_filters3d_smp
  
  subroutine build_filters3d_smp2(filt,ncoef)
    
    type(nfilter)        :: filt   ! filter
    integer              :: ncoef
    
    call build_cartesian_3d_space_smp(filt,npatch,ncoef)
    
  end subroutine build_filters3d_smp2
  
  subroutine build_cartesian_3d_space_smp(filt,npatch,ncoef)
    
    type(nfilter)        :: filt
    integer              :: npatch(3),ncoef
    integer              :: i1,i2,i3,i,iref,iout(3),j(3)
    integer, allocatable :: map(:)
    
    allocate(map(product(n)))
    
    j(1) = ceiling(real(n(1))/real(npatch(1))) ! number of coefficients
    j(2) = ceiling(real(n(2))/real(npatch(2))) ! per filter in the
    j(3) = ceiling(real(n(3))/real(npatch(3))) ! 3 directions
    
    do i3=1,n(3)
       iout(3)=ceiling(real(i3)/real(j(3)))
       do i2=1,n(2)
          iout(2)=ceiling(real(i2)/real(j(2)))
          do i1=1,n(1)
             iout(1)=ceiling(real(i1)/real(j(1)))
             i=i1+n(1)*(i2-1)+n(1)*n(2)*(i3-1)
             iref=iout(1)+npatch(1)*(iout(2)-1)+ &
                  &npatch(2)*npatch(1)*(iout(3)-1)
             map(i)=iref
          end do
       end do
    end do

    filt = createnhelix_smp(n,center,(/0,0,0/),na,map,nmis)
    ncoef= size(filt%hlx(1)%lag)
    
    deallocate(map)
    
  end subroutine build_cartesian_3d_space_smp

  subroutine build_cartesian_3d_space(filt,npatch,ncoef)
    
    type(nfilter)        :: filt
    integer              :: npatch(3),ncoef
    integer              :: i1,i2,i3,i,iref,iout(3),j(3)
    integer, allocatable :: map(:)
    
    allocate(map(product(n)))
    
    j(1) = ceiling(real(n(1))/real(npatch(1))) ! number of coefficients
    j(2) = ceiling(real(n(2))/real(npatch(2))) ! per filter in the
    j(3) = ceiling(real(n(3))/real(npatch(3))) ! 3 directions
    
    do i3=1,n(3)
       iout(3)=ceiling(real(i3)/real(j(3)))
       do i2=1,n(2)
          iout(2)=ceiling(real(i2)/real(j(2)))
          do i1=1,n(1)
             iout(1)=ceiling(real(i1)/real(j(1)))
             i=i1+n(1)*(i2-1)+n(1)*n(2)*(i3-1)
             iref=iout(1)+npatch(1)*(iout(2)-1)+ &
                  &npatch(2)*npatch(1)*(iout(3)-1)
             map(i)=iref
          end do
       end do
    end do

    filt = createnhelix(n,center,(/0,0,0/),na,map)
    ncoef= size(filt%hlx(1)%lag)
    
    deallocate(map)
    
  end subroutine build_cartesian_3d_space

  subroutine create_lap_3d_rough(rough,npatch,ncoef)
    
    type(nfilter)        :: rough
    integer              :: ncoef,npatch(3)
    integer              :: npts(3)
    integer, allocatable :: nh(:),map(:)
    
    npts=(/npatch(1),npatch(2),npatch(3)/)
    allocate(nh(product(npts)))
    allocate(map(ncoef*product(npts)))
    
    nh=ncoef
    map=1
    
    nh=min(6,npts(1))*3
    call nallocate(rough,nh,map)
    rough%hlx(1)=lapfac3d(.01,npts(1),npts(2),nh(1)/3)
        
    deallocate(nh,map)
    
  end subroutine create_lap_3d_rough
  
  subroutine build_filter3d_close()
    deallocate(nmis)
  end subroutine build_filter3d_close

end module build_filter3d














