module findmatch_mod
  
  use helix
  use helicon
  use polydiv
  use nchconest
  use nchelicon
  
  use nhelix
  use ncnhconest
  use ncnhelicon
  
  use cgstep_mod
  use solver2_mod
  use verbosity_mod
  
  use mkwallwt
  use patch
  use tent
  use random

  ! Weight = L1
  use irls
  use irls_priv

  ! Weight = PEF
  use pef_pat
  use bound
  use print
  use compress
  use misinput
  use createhelixmod
  use solver_smp_mod

  ! NS pef
  use nhelicon
  use solver_smpPEF_mod
  use solver_regPEF_mod
  use solver_prcPEF_mod

  implicit none
  
  integer, private                     :: nf,ndata
  type (filter), private               :: acorr
  real, private, dimension(:), pointer :: xcorr,space,matrix

contains

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !! Find a single matched-filter     !!
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  ! Minimizes residual:
  ! wght ( data * match - targt ) = 0
  subroutine findmatch(data,targt,match,niter,init,noise,nmem_in,nfreq_in)
    optional                                           ::nmem_in,nfreq_in
    real, dimension(:), pointer           :: data,targt
    integer                               :: niter,nmem_in,nfreq_in
    logical,                     optional :: init
    real,                        optional :: noise
    type (filter) :: match
    integer       :: stat,i,nmem,nfreq
    logical       :: cginv,initchol
    real          :: sd,rnd

    cginv=(niter.ge.0)
 
    if (present(nfreq_in)) then 
       nmem=nmem_in
       nfreq=nfreq_in
    end if

    if (present(noise)) then
       sd=noise*sqrt(dot_product(data,data))
       do i=1,size(data)
          rnd=sd*random_normal()
          data(i)=data(i)+rnd
       end do
    end if

    if (present(nmem_in)) then
       call cgfindmatch(data,targt,match,niter,nmem,nfreq)
    else
       call cgfindmatch(data,targt,match,niter)
    end if
  end subroutine findmatch

  !! Find match with conjugate gradients (slow)
  subroutine cgfindmatch(data,targt,match,niter,nmem,nfreq)
    optional                                 :: nmem,nfreq
    real, dimension(:), pointer           :: data,targt
    integer                               :: niter,nmem,nfreq
    type (filter) :: match
    integer       :: stat,niter

    if (verb>1) write(0,*) ' Using CG to solve system- there may be better way'
    call nchconest_init(data,match)
    if (present(nfreq)) then
       write(0,*) 'Using L1 norm with nonlinear CG'
       write(0,*) 'nmem',nmem
       write(0,*) 'nfreq',nfreq
       call solver(nchconest_lop,cgstep,match%flt,targt,niter,wght=l1_priv, &
            &     nmem=nmem,nfreq=nfreq)
    else
       call solver(nchconest_lop,cgstep,match%flt,targt,niter)
    end if

    call cgstep_close()

  end subroutine cgfindmatch

  subroutine findmatchPEF(data,targt,match,niter,nfpef,ndim,n,init,noise)
    real,    dimension(:), pointer        :: data,targt
    integer, dimension(:), pointer        :: nfpef,n
    integer                               :: niter
    integer                               :: ndim                               
    logical,                     optional :: init
    real,                        optional :: noise
    type (filter)                         :: match
    integer                               :: stat,i,nmem,nfreq
    logical                               :: cginv,initchol
    real                                  :: sd,rnd

    cginv=(niter.ge.0)   

    if (present(noise)) then
       sd=noise*sqrt(dot_product(data,data))
       do i=1,size(data)
          rnd=sd*random_normal()
          data(i)=data(i)+rnd
       end do
    end if

    call cgfindmatchPEF(data,targt,match,niter,nfpef,ndim,n)
   
  end subroutine findmatchPEF

  !! Find match with conjugate gradients (slow) for the PEF
  subroutine cgfindmatchPEF(data,targt,match,niter,nfpef,ndim,n)
    
    real,    dimension(:), pointer     :: data,signal,targt,kk
    integer, dimension(:), pointer     :: nfpef,n
    integer                            :: niter,ndim
    type (filter)                      :: match, Np, Sp
    integer                            :: stat,niter
    integer, dimension(:), allocatable :: n0,gap,center
    

    allocate(n0(ndim), center(ndim), gap(ndim))
    allocate(kk(product(n)))
    allocate(signal(size(targt)))
    
    gap    = 0
    n0     = n
    center = 1
    kk     = 1.

    where(nfpef(2:)>1) center (:ndim-1) = nfpef(:ndim-1)/2 + 1
    where(nfpef(2:)>1) center (:ndim-1) = nfpef(:ndim-1)/2 + 1
    Np =  createhelix( n, center, gap, nfpef)
    Sp =  createhelix( n, center, gap, nfpef)

    call boundn   ( n0, n, nfpef, Np)
    call boundn   ( n0, n, nfpef, Sp)

    call find_mask((kk /= 0.), Np)
    call find_mask((kk /= 0.), Sp)
    
    where (Np%mis) 
       kk = 0.
    elsewhere
       kk = 1.
    end where  
    
    call find_pef_pat(data, Np, 2*size(Np%lag)) ! Estimate noise PEF N
    call resize  (Np)
    
    write(0,*) '######################'
    write(0,*) '#                    #'
    write(0,*) '#     Noise  Pef     #'
    write(0,*) '#                    #'
    write(0,*) '######################'
    
    call printn (n, center, nfpef, Np)
    call helicon_init (Np)

    stat = helicon_lop(.false.,.false.,targt,signal) ! Compute Nd
    signal=signal*kk
    call find_pef_pat(signal, Sp, 2*size(Sp%lag)) ! Estimate signal PEF S
    call resize  (Sp)
    
    write(0,*) '######################'
    write(0,*) '#                    #'
    write(0,*) '#     Signal  Pef    #'
    write(0,*) '#                    #'
    write(0,*) '######################'

    call printn       (n, center, nfpef, Sp)
    call helicon_init (Sp)

    if (verb>1) write(0,*) ' Using CG to solve system- there may be better way'
    
    call nchconest_init(data,match)
    call solver_smp2( m=match%flt, d=targt, Fop=nchconest_lop, &
    & stepper=cgstep, niter=niter, Wop=helicon_lop, wt=kk)
    
    deallocate(n0, center, gap, kk, signal)
    call cgstep_close()

  end subroutine cgfindmatchPEF

  !! Calculates:
  !! output = input * match
  subroutine applymatch(input,output,match)
    
    real, dimension(:), pointer :: input, output
    type (filter)               :: match
    integer stat
    
    call nchelicon_init(match)
    stat=nchelicon_lop(.false.,.false.,input,output)
 
  end subroutine applymatch

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !! Non-stationary matched filtering !!
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  !! Finds non-stationary matched-filter
  subroutine findnsmatch(data,targt,match,reg,mode,niter,eps,wght,nfreq_in,nmem_in)
    real, dimension(:), pointer           :: data,targt
    type (nfilter)                        :: match
    type ( filter)                        :: reg
    integer                               :: mode
    integer, optional                     :: niter,nfreq_in,nmem_in
    real, optional                        :: eps
    real, dimension(:), pointer, optional :: wght

    real, dimension(:), pointer :: coeffs
    integer :: stat,nhmax,nitr,i,nfreq,nmem
    logical :: sverb
    real    :: epsilon

    sverb=(verb>2)
    nhmax=size(match%hlx(1)%lag)  ! assumes all filters are same length
    if (present(wght)) call seperr('Weight not yet implemented in findnsmatch')

    nitr=size(match%hlx(1)%lag)-1
    if (present(niter)) nitr=niter
    epsilon=1.
    if (present(eps)) epsilon=eps

    write(0,*) 'epsilon = ',epsilon

    allocate (coeffs(nhmax*size(match%hlx)))
    coeffs=0.

    select case(mode)

    case (1)  ! Regularized solver

       if (verb>2) write(0,*) '  using regularized conjugate gradients'
       call helicon_init(reg)                       ! initialize regularizer
       call ncnhconest_init(data,match,nhmax)       ! initialize hconest

       nfreq=-1
       nmem=5
       if (present(nfreq_in)) nfreq=nfreq_in
       if (present(nmem_in)) nmem=nmem_in

       ! solve problem
       if (nfreq>0) then
          write(0,*) 'nfreq',nfreq,'nmem',nmem
          call solver_reg(ncnhconest_lop,cgstep,helicon_lop,size(coeffs), &
          &   coeffs,targt,nitr,epsilon,x0=coeffs,nmem=nmem, &
          &   nfreq=nfreq,wght=l1_priv,verb=sverb)
       else
          call solver_reg(ncnhconest_lop,cgstep,helicon_lop,size(coeffs), &
          &   coeffs,targt,nitr,epsilon,x0=coeffs,verb=sverb)
       end if

    case (2)  ! Preconditioned solver

       if (verb>2) write(0,*) '  using preconditioned conjugate gradients'
       call polydiv_init(size(coeffs),reg)          ! initialize preconditioner
       call ncnhconest_init(data,match,nhmax)       ! initialize hconest

       nfreq=-1
       nmem=5
       if (present(nfreq_in)) nfreq=nfreq_in
       if (present(nmem_in)) nmem=nmem_in

       ! solve problem
       if (nfreq>0) then
          write(0,*) 'nfreq',nfreq,'nmem',nmem
          call solver_prec(ncnhconest_lop,cgstep,polydiv_lop,size(coeffs), &
          &   coeffs,targt,nitr,epsilon,x0=coeffs,nmem=nmem, &
          &   nfreq=nfreq,wght=l1_priv,verb=sverb)
       else
          call solver_prec(ncnhconest_lop,cgstep,polydiv_lop,size(coeffs),coeffs,targt,nitr,epsilon,x0=coeffs,verb=sverb)
       end if

    case (3)  ! Simple solver

       if (verb>2) write(0,*) '  CG with no regularization - mileage may vary'
       call ncnhconest_init(data,match,nhmax)       ! initialize hconest
       call solver(ncnhconest_lop,cgstep,coeffs,targt,nitr,x0=coeffs,verb=sverb)

    case default
       call seperr('Mode not implemented in findnsmatch')
    end select

    if (verb>2) write(0,*) '  Done with solver'

    !! Copy coefficients into NS filter type
    do i=1,size(match%hlx)
       match%hlx(i)%flt(1:nhmax)=coeffs(1+(i-1)*nhmax:i*nhmax)
    end do

    deallocate(coeffs)
    call cgstep_close()

  end subroutine findnsmatch
  
  subroutine findnsmatchPEF(data,targt,match,sigpef,reg,mode,niter,eps)
    real, dimension(:), pointer           :: data,targt
    type (nfilter)                        :: match
    type (nfilter),     pointer           :: sigpef
    type ( filter)                        :: reg
    integer                               :: mode
    integer, optional                     :: niter
    real,    optional                     :: eps
    real, dimension(:), pointer           :: coeffs
    integer                               :: stat,nhmax,nitr,i
    logical                               :: sverb
    real                                  :: epsilon

    sverb=(verb>2)
    nhmax=size(match%hlx(1)%lag)  ! assumes all filters are same length
    nitr=size(match%hlx(1)%lag)-1

    if (present(niter)) nitr=niter
    epsilon=1.
    if (present(eps)) epsilon=eps

    write(0,*) 'epsilon = ',epsilon

    allocate (coeffs(nhmax*size(match%hlx)))
    coeffs=0.

    call nhelicon_init(sigpef)
    call ncnhconest_init(data,match,nhmax)

    select case(mode)

    case (1)  ! Regularized solver

       if (verb>2) write(0,*) '  using regularized conjugate gradients with PEF'
       call helicon_init(reg)                       ! initialize regularizer
       call solver_regPEF(m=coeffs,d=targt,Fop=ncnhconest_lop,Aop=helicon_lop,&
       &   stepper=cgstep, nAop=size(coeffs),niter=nitr,eps=epsilon,m0=coeffs,&
       &   verb=sverb,Wop=nhelicon_lop)
       call helicon_close()

    case (2)  ! Preconditioned solver

       if (verb>2) write(0,*) '  using preconditioned conjugate gradients with PEF'
       call polydiv_init(size(coeffs),reg)          ! initialize preconditioner
       call solver_prcPEF(m=coeffs,d=targt,Fop=ncnhconest_lop,Sop=polydiv_lop,&
       &   stepper=cgstep,nSop=size(coeffs),niter=nitr,eps=epsilon,&
       &   p0=coeffs,verb=sverb,Wop=nhelicon_lop)
       call polydiv_close()

    case (3)  ! Simple solver

       if (verb>2) write(0,*) '  CG with no regularization with PEF - mileage may vary'
       call solver_smp2(m=coeffs, d=targt, Fop=ncnhconest_lop, &
       & stepper=cgstep, niter=nitr, Wop=helicon_lop, m0=coeffs, verb=sverb)

    case default
       call seperr('Mode not implemented in findnsmatch')
    end select

    if (verb>2) write(0,*) '  Done with solver'
    
    call cgstep_close()
    call nhelicon_close()
    call ncnhconest_close()
    
    !! Copy coefficients into NS filter type
    do i=1,size(match%hlx)
       match%hlx(i)%flt(1:nhmax)=coeffs(1+(i-1)*nhmax:i*nhmax)
    end do
    deallocate(coeffs)
  end subroutine findnsmatchPEF
  
  !! Applies non-stationary matched-filter
  subroutine applynsmatch(input,output,match)
    real, dimension(:), pointer :: input, output
    type (nfilter)              :: match
    integer stat
    call ncnhelicon_init(match)
    stat=ncnhelicon_lop(.false.,.false.,input,output)
    call ncnhelicon_close()
  end subroutine applynsmatch

  !! Flags edges of cube as missing data
  !! Eliminates helical wrap around
  subroutine flagedges(missing,nwind)
    logical, dimension(:), pointer :: missing
    integer, dimension(:), pointer :: nwind
    integer :: i,j,k
    k=1
    missing=.false.
    do i=1,size(nwind)-1
       k=k*nwind(i)
       do j=1,nwind(i+1)
          missing(j*k)=.true.
       end do
    end do
  end subroutine flagedges
  
end module findmatch_mod












