! Copyright 1999 Stanford University
!All rights reserved
!
!Author: Sergey Fomel and Robert Clapp
module npef_pat
  ! Estimate non-stationary PEF
  use nhconest
  use npolydiv
  use polydiv
  use cgstep_mod
  use solver2_mod
  use dottest
  use hconest
  use nhelicon
  use wilson
  use autocorr
  use helix
  implicit none
contains
  subroutine find_pef( dd, aa, rr, niter, eps, nh, wt, verb)
    optional                         :: wt
    integer,             intent(in)  :: niter, nh      ! number of iterations
    real,                intent(in)  :: eps	       ! epsilon
    type( nfilter)                   :: aa             ! estimated PEF output.
    type( nfilter),      intent(in)  :: rr             ! roughening filter.
    type(  filter)                   :: dummy
    real, dimension(:),     pointer  :: dd             ! input data
    real, dimension(:,:),   pointer  :: rmov         ! input data
    integer                          :: ip, ih, np, nr ! filter lengths
    real, dimension (:), allocatable :: flt	       ! np*na filter coefs
!    real, dimension (:), allocatable :: temp1,temp2
    real, dimension (:)              :: wt
    real :: dot1(2),dot2(2),scale,tempr
    logical :: verb
    integer :: i1,ierr

    np = size( aa%hlx)			! data   length
    nr = np*nh
    allocate( flt( nr))
    flt=0.

    call nhconest_init( dd, aa, nh)
!    call npolydiv_init( nr, rr)

    dd=dd-sum(dd)/size(dd)

!    if (present(wt)) then
!       call solver_prec(nhconest_lop,cgstep,npolydiv_lop,nr,flt,-dd, &
!       & niter,eps,wt=wt,verb=verb)
!    else
!       call solver_prec(nhconest_lop,cgstep,npolydiv_lop,nr,flt,-dd, &
!       & niter,eps,verb=verb)
!    end if
!
    call nhelicon_init(rr)
    
    if (present(wt)) then
       call solver_reg(nhconest_lop,cgstep,nhelicon_lop,nreg=nr, &
       & x=flt,dat=-dd,niter=niter,eps=eps,verb=verb,wt=wt)
    else
       call solver_reg(nhconest_lop,cgstep,nhelicon_lop,nreg=nr, &
       & x=flt,dat=-dd,niter=niter,eps=eps,verb=verb)
    end if
    
    call   cgstep_close()
    call nhelicon_close()    
!   call npolydiv_close()
    call nhconest_close()

    do ip = 1, np 
       do ih = 1, size( aa%hlx(ip)%lag)
          aa%hlx( ip)%flt( ih) = flt( (ip-1)*nh + ih)
       end do
    end do

    deallocate(flt)
    return

  end subroutine find_pef

end module npef_pat














