module npolydiv2  
! Helix polynomial division
  use adj_mod
  use nhelix
  implicit none
  integer, private :: nd 
  type( nfilter), private,save :: aa 
  real ,private,  dimension (:), allocatable  :: tt 
  contains
!----------------------------------------------------------------------
  subroutine npolydiv2_init ( nd_in,aa_in )
    integer    :: nd_in 
    type( nfilter)    :: aa_in 
    nd = nd_in 
    aa = aa_in 
    if (.not. allocated(tt)) then
      allocate(tt ( nd)) 
    end if
  end subroutine npolydiv2_init
!----------------------------------------------------------------------
  function npolydiv2_lop ( adj, add, xx, yy) result(stat)
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: xx,yy 
    call adjnull (adj,add,xx,yy )
    call npolydiv2_lop2(adj,xx,yy )
    stat=0
  end function npolydiv2_lop
!----------------------------------------------------------------------
  subroutine npolydiv2_lop2(adj,xx,yy)
    logical,intent(in) :: adj 
    real, dimension (:)  :: xx 
    real, dimension (:)  :: yy 
    integer                        :: ia, ix, iy, ip
    integer, dimension(:), pointer :: lag
    real,    dimension(:), pointer :: flt

    tt = 0.
    if ( adj) then
      tt = yy
      do iy= nd, 1, -1  
        ip = aa%pch( iy)
        lag => aa%hlx( ip)%lag
        flt => aa%hlx( ip)%flt
        do ia = 1, size( lag)
          ix = iy - lag( ia)
          if ( ix < 1) then
            cycle
          end if
          tt( ix) =			tt( ix) -  flt( ia) * tt( iy)
        end do
      end do 
      xx =	xx + tt
    else
      tt = xx
      do iy= 1, nd  
        ip = aa%pch( iy)
        lag => aa%hlx( ip)%lag
        flt => aa%hlx( ip)%flt
        do ia = 1, size( lag)
          ix = iy - lag( ia)
          if ( ix < 1) then
            cycle
          end if
          tt( iy) =			tt( iy) -  flt( ia) * tt( ix)
        end do
      end do 
      yy =	yy + tt
    end if
  end subroutine npolydiv2_lop2
!----------------------------------------------------------------------
  subroutine npolydiv2_close()
    deallocate( tt )
  end subroutine npolydiv2_close
!----------------------------------------------------------------------
end module npolydiv2



