! Copyright 1999 Stanford University
!All rights reserved
!
!Author:  Robert Clapp / Antoine Guitton for Nemeth thing

module nsignoi  

  use adj_mod
  use nhelicon
  use npolydiv
  use npolydiv2
  use nhelicon2
  use cgstep_mod
  use solver2_mod
  use hell_row_mod
  use mat_row_init

  implicit none

  integer,   private :: i 
  integer,   private :: niter 
  integer,   private :: nd 
  integer,   private :: separmod 
  real,      private :: mu
  logical,   private :: verb 
  real,      private :: eps 
  type( nfilter) ,private,   pointer              :: nn 
  type( nfilter) ,private,   pointer              :: ss 
!  real ,private,  dimension (:), allocatable      :: dd 
  
contains

  subroutine nsignoi_init ( nn_in,ss_in,niter_in,nd_in,eps_in,separmod_in&
  &,mu_in,verb_in )
    
    type( nfilter) ,    pointer :: nn_in 
    type( nfilter) ,    pointer :: ss_in 
    integer                     :: niter_in 
    integer                     :: nd_in 
    real                        :: eps_in 
    real                        :: mu_in 
    integer                     :: separmod_in 
    logical                     :: verb_in 
    
    nn       => nn_in 
    ss       => ss_in 
    niter    = niter_in 
    nd       = nd_in 
    eps      = eps_in 
    mu       = mu_in 
    separmod = separmod_in 
    verb     = verb_in 
    
!    if (.not. allocated(dd)) then
!      allocate(dd ( nd)) 
!    end if
  
  end subroutine nsignoi_init
  
  function nsignoi_lop ( adj, add, data, signal, wt) result(stat)
    
    integer            :: stat 
    logical,intent(in) :: adj,add 
    real,dimension(:)  :: data,signal,wt
    call adjnull (adj,add,data,signal)
    call nsignoi_lop2(adj,data,signal,wt)
    
    stat=0
  
  end function nsignoi_lop
  
  subroutine nsignoi_lop2(adj,data,signal,wt)
  
    logical,intent(in)   :: adj 
    real, dimension (:)  :: data 
    real, dimension (:)  :: signal,wt
    integer              :: stat1
    integer,external     :: getch
    character(len=128)   :: temp_ch
!    real, dimension (2*size(data)),target :: m
    real, dimension (:), allocatable,target :: m
    real, dimension (:),pointer :: mo, no  ! model for signal
    real, dimension (:), allocatable :: dd
    
    if (separmod.eq.0) then
       
       allocate(m(2*size(data)))
       no => m(1:size(data)); mo => m(1+size(data):)
       
       if (verb) write(0,*) 'SEPAR !----------------------------'
       if (verb) write(0,*) 'SEPAR ! Nemeth method Lsms+Lnmn=d  '
       if (verb) write(0,*) 'SEPAR !----------------------------'

       call npolydiv_init (nd, ss)       
       call npolydiv2_init(nd, nn)

       call init_hell_row(nd,mu)
       call hell_mat_init(nd,nd)

       if (eps.eq.0.) then

          call solver( hell_row_op, cgstep, niter=niter, x=m ,  &
               & dat=data, verb=verb, wt=wt)
       
       else
          
          call solver_reg( hell_row_op, cgstep, hell_mat_op, &
               & nreg=2*nd, niter=niter,  x=m ,  &
               & dat=data, verb=verb, eps=eps, wt=wt)

       end if

       stat1 = npolydiv_lop(.false.,.false.,mo,signal)
       call npolydiv_close()
       call npolydiv2_close()
       deallocate(m)
    else

       if (verb) write(0,*) 'SEPAR !---------------------------- '
       if (verb) write(0,*) 'SEPAR !  Soubaras method N(s-d)=0   '
       if (verb) write(0,*) 'SEPAR !---------------------------- '       
       
       call nhelicon_init (nn)
       call nhelicon2_init(ss,wt)
       allocate(dd(size(data)))
       dd=data
       stat1=nhelicon_lop(.false.,.false.,dd,data)  !Nd
       deallocate(dd)
       call solver_reg (nhelicon_lop, cgstep, nhelicon2_lop, &
       & nreg=nd, niter= niter, x=signal,&
       & dat=data, eps=eps, verb=verb, wt=wt)
       call nhelicon_close()
       call nhelicon2_close()
       
    end if

    call cgstep_close()
    
  end subroutine nsignoi_lop2
  
  subroutine  nsignoi_close()
!    if (allocated(dd)) then
!       deallocate(dd) 
!    end if
  end subroutine nsignoi_close

end module nsignoi
































