! Convo.x <in.H >out.H
!
!   shot=1 if one-sided gather
!   shot=0 if two-sided gather
!   Work fine with 2-D data.
!   I did not tested it with 3-D (use at your on risk).

program convo 
  use sep
  implicit none

  integer :: n1,n2,n3,i3
  integer :: n123,n123pad,n12
  integer :: jpad2
  integer :: shot
  real    :: d1,d2,o2
  real, allocatable, dimension(:)    :: data
  complex, allocatable, dimension(:) :: cdata

  call sep_init()
  call from_par('shot',shot,1)
  call from_history(n1,n2,n3)
  call from_history('d1',d1)
  call from_history('d2',d2)
  call from_history('o2',o2)
  call to_history('n2',n2)
  call sep_close()

  n123=2*n1*n2*n3
  n12=n1*n2
  n123pad=jpad2(n123)
  write(0,*) 'n123=',n123,'   (',real(n123pad*8)/real(1024.*1024.),'Mbytes)'

  allocate(data(n12),cdata(n123pad))

  write(0,*) '  reading data...'
  cdata=0.
  do i3=1,n3
     call sep_read(data)
     cdata((i3-1)*n12+1:(i3-1)*n12+n12)=cmplx(data(1:n12),0.)
  end do
  write(0,*) '  calling ftu...'
  call jftu( 1., n123pad, cdata)
  cdata=cdata*cdata

  write(0,*) '  calling inverse ftu...'
  call jftu(-1., n123pad, cdata)

  do i3=1,n3
     if (shot==0) then                      ! two-sided gather
        data(1:n1*n2)=cdata((i3-1)*n3+n1*(n2/2)+1:(i3-1)*n3+n1*(n2/2))
     else if (abs(o2+d2).lt.abs(o2)) then   ! decreasing offset
        data(1:n1*n2)=cdata((i3-1)*n3+n1*n2+1:(i3-1)*n3+2*n1*n2)
     else                                   ! increasing offset
        data(1:n1*n2)=cdata((i3-1)*n3+1:(i3-1)*n3+n1*n2)     
     end if
     call sep_write(data)
  end do

end program convo 

subroutine jftu( signi, nx, cx )
!   complex fast fourier transform with unitary scaling
!
!               1         nx          signi*2*pi*i*(j-1)*(k-1)/nx
!   cx(k)  =  -------- * sum cx(j) * e
!             sqrt(nx)   j=1             for k=1,2,...,nx=2**integer
!
  integer :: nx, i, j, k, m, istep, jpad2
  real    :: signi
  real*8  :: arg,pi
  complex :: cx(nx), cw, cdel, ct
  pi=4.0d0*datan(1.0d0)
  if( nx /= jpad2(nx) )  call erexit('ftu: nx not a power of 2')
  cx = cx/sqrt( 1.*nx)
  j = 1;  k = 1
  do i= 1, nx
     if (i<=j) then
        ct = cx(j); cx(j) = cx(i); cx(i) = ct
     end if
     m = nx/2
     do while (j>m .and. m>1) 
        j = j-m; m = m/2 
     end do
     j = j+m
  end do
  do 
     istep = 2*k;   cw = 1.;   arg = signi*pi/k
     cdel = cmplx( cos(arg), sin(arg))
     do m= 1, k
        do i= m, nx, istep
           ct=cw*cx(i+k);  cx(i+k)=cx(i)-ct;  cx(i)=cx(i)+ct
        end do
        cw = cw * cdel
     end do
     k = istep
     if (k>=nx) exit
  end do
end subroutine jftu

integer function jpad2(n)
  integer n
  jpad2=1
  do
     if (jpad2.ge.n) exit
     jpad2=jpad2*2
  end do
end function jpad2
