program Mute_pick

  use sep

  implicit none

  integer nt,nx,  n3
  integer trcbal, weight, taper
  real tramp, xpow
  real eramp, emute
  !real t0, dt,  x0, dx, tmute0, vmute, tpow
  real t0, dt,  x0, dx, vmute, tpow
  real, allocatable, dimension(:) :: trace,scale
  real, allocatable, dimension(:) :: tmute,tmute0
  integer,  allocatable, dimension(:) :: kill
  logical,  allocatable, dimension(:) :: lkill
  integer, dimension(:), pointer :: n 
  real, dimension(:), pointer :: d,o
  integer                     :: ndim

  call sep_init(SOURCE)
  ndim=sep_dimension()

  allocate(n(ndim),d(ndim),o(ndim))
  call from_history('n',n)
  call from_history('o',o)
  call from_history('d',d)

  t0=o(1)
  x0=o(2)
  dt=d(1)
  dx=d(2)
  nt=n(1)
  nx=n(2)

  n3=max(n(3),1)

  call from_param('vmute',vmute,1450.)
  call from_param('tramp',tramp,0.25)
  call from_param('tpow',tpow,0.)
  call from_param('xpow',xpow,0.)
  call from_param('trcbal',trcbal,0)
  call from_param('weight',weight,0)
  call from_param('taper',taper,0)
  emute=t0+nt*dt
  call from_param('eramp',eramp,0.25)
  call from_param('emute',emute,emute)


  allocate(trace(nt),scale(nt),tmute(nt),tmute0(nt),kill(nx),lkill(nx))

 call muted(t0,x0,dt,dx,nt,nx,n3,vmute,tramp,tpow,xpow,trcbal,weight&
  &,taper,trace,scale,kill,lkill,emute,eramp)

contains

  subroutine muted(t0,x0,dt,dx,nt,nx,n3,vmute,tramp,tpow,xpow,trcbal,weight&
  &,taper,trace,scale,kill,lkill,emute,eramp)

    real t0,x0,dt,dx,vmute,tramp,tpow,xpow,emute,eramp
    integer nt,nx,n3,trcbal,weight,taper
    integer nread,it,ix,i3
    integer sreed
    real tmute(nt), tmute0(nt)
    real trace(nt),scale(nt)
    logical lkill(nx)
    integer kill(nx),nkill,fetch
    !real tpower, tmute, xpower, xratio, eps
    real tpower, xpower, xratio, eps
    real t, x, w, r0, r,drop,xmin,xmax
    integer hetch,putch
    real ptmute,ptmute0
    integer mread, ptime

    ! picked mute values

    nkill = 0

    if (nkill > 0) then
       do ix=1,nx
          lkill(ix) = .false.
       end do
       do ix=1,nkill
          lkill(kill(ix)) = .true.
       end do
    end if
    
    ! Initializations...
    eps= 1.0e-12
    drop = 10.
    r0 = sqrt( sqrt( drop -1. ) )
    xmin = min( abs(x0), abs(x0+(nx-1)*dx) )
    xmax = max( abs(x0), abs(x0+(nx-1)*dx) )
    ! Main loops...
    do i3 = 1, n3
       ptmute = 0.
       ptmute0 = 0.
!       call sep_read('pick',tmute)
       mread  = sreed( 'pick', tmute, 4*nt)
       if ( mread .ne. 4*nt ) then
          call erexit('read failed for pick function')
       end if
       call find0(nt,tmute,ptime)
       ptmute=t0+ptime*dt
       ptmute0 = ptmute
       do ix = 1, nx  
          x = x0 + (ix-1)*dx
          ! Read a trace...
          nread = sreed( 'in', trace, 4 * nt )
          if ( nread .ne. 4*nt ) then
             call erexit('read failed')
          end if
          ! handle trace kills
          if (nkill .ne. 0) then
             if (lkill(ix)) then
                do it=1,nt
                   trace(it) = 0.0
                end do
                call srite ('out' , trace, 4*nt )
                !
                cycle
                !
             end if
          end if
          ! Scale trace by tpow...
          if ( tpow .ne. 0 ) then
             do it = 1, nt  
                tpower = (t0 + (it-1) * dt) ** tpow 
                trace(it) = trace(it) * tpower
             end do
          end if
          ! Do trace balancing...
          if ( trcbal .ne. 0 ) then
             if ( trcbal .eq. 1 ) then
                call absnorm( nt, trace)
             end if
             if ( trcbal .eq. 2 ) then
                call rmsnorm( nt, trace)
             end if
             if ( trcbal .eq. 3 ) then
                call e2norm( nt, trace)
             end if
          end if
          ! Scale trace by xpow...
          if ( xpow .ne. 0 ) then
             xratio = x/(xmin + eps)
             xpower = xratio ** xpow
             do it = 1, nt  
                trace(it) = trace(it) * xpower
             end do
          end if
          ! Calculate and apply mute scale mask...
          ptmute = sqrt(max(0.0,ptmute0**2 + (x/vmute)**2))
          do it = 1, nt  
             t = t0 + (it-1)*dt 
             ! front mute
             if ( t <= ptmute ) then
                scale(it) = 0.
             else if ( t <= ptmute+tramp ) then
                scale(it) = (t-ptmute)/(tramp+1.e-20)
             else
                scale(it) = 1.
             end if
             ! tail mute
             if ( t >= emute+eramp ) then
                scale(it) = 0.
             else
                if (t > emute) then
                   scale(it) = scale(it)*(1.0 - (t-emute)/(eramp+1.e-20))
                end if
             end if
             trace(it) = scale(it)*trace(it)
          end do
          ! Do Jon's weighting...
          if ( weight .ne. 0) then
             do it = 1, nt  
                trace(it) = trace(it) / ( t0 + (it-1)*dt + xmax / vmute )
             end do
             x = x0 + (ix-1) * dx
             x = abs(x)
             do it = 1, nt  
                t = t0 + it * dt
                r =  r0 * x / ( vmute * t + 1.e-20 )
                w =  r / ( 1. + (r*r)**2 )
                ! denominator is 1/drop at cutoff.
                trace(it) = trace(it) * w
             end do
          end if
          ! Do gather boundary tapers...
          if ( taper .ne. 0 ) then
             if ( ix <= taper ) then
                do it = 1, nt  
                   trace(it) = trace(it) * (ix-1)/(1.*taper)
                end do
             else if ( ix >= nx-taper+1 ) then
                do it = 1, nt  
                   trace(it) = trace(it) * (nx - ix)/(1.*taper)
                end do
             end if
          end if
          ! Write out the muted trace...
          call srite( 'out', trace,  nt * 4 )
       end do
    end do
  end subroutine muted

  subroutine absnorm( n, x )
    integer  i, n
    real x(n), avg
    avg = 1.e-30
    do i = 1, n
       avg = avg + abs(x(i))
    end do
    avg = avg / n
    do i = 1, n
       x(i) = x(i) / avg
    end do
    return
  end subroutine absnorm
  subroutine rmsnorm( n, x )
    integer  i, n
    real x(n), avg
    avg = 1.e-30
    do i = 1, n
       avg = avg + x(i)*x(i)
    end do
    avg = sqrt(avg) / n
    do i = 1, n
       x(i) = x(i) / avg
    end do
    return
  end subroutine rmsnorm
  subroutine e2norm( n, x )
    integer  i, n
    real x(n), avg
    avg = 1.e-30
    do i = 1, n
       avg = avg + x(i)*x(i)
    end do
    avg = avg / n
    do i = 1, n
       x(i) = x(i) / avg
    end do
    return
  end subroutine e2norm
  subroutine find0( n, x, p)
    integer n, i, p
    real x(n)
    i=1
    do while  (x(i).eq.0)
       i=i+1
    end do
    p=i-1
    return
  end subroutine find0

end program Mute_pick
