!  Noah-conv.x
!
!  Creates a multiple model for a 2D line. No amplitude
!  term has been implemented. Good for adaptive subtraction only !!
!  
!  Input:   Shot Gathers in (w,x), sqrt(t) corrected
!
!           n1 offset
!           n2 shot
!           n3 frequency
!
!   Make sure d1==d2 or distance between two shots == receiver separation
!   Make sure that h=g-s and not s-g
!   Make sure that ns=1 corresponds to the last shot
!   
!   For more details, read Dragoset/Jericevic 1998 vol 63 Geophysics p.772-789
!
!%

program noah_conv

  use sep
  implicit none

  integer :: nw,ns,nh,iw,is,ih,ihp,esize,ih0,nhits
  real    :: w,w0,pi,oh,dh,dw
  complex :: weight
  complex, allocatable, dimension(:,:) :: u0,u1

  call sep_init(SOURCE)
  call from_history('n1',nh)
  call from_history('d1',dh)
  call from_history('o1',oh)
  call from_history('n2',ns)
  call from_history('n3',nw)
  call from_history('o3',w0)
  call from_history('d3',dw)
  call from_history('esize',esize)
  if (esize.ne.8) call seperr('esize.ne.8\n')

  pi=3.14159265

  allocate(u0(nh,ns))
  allocate(u1(nh,ns))

  call sep_close()

  ih0=int(oh/dh)+1
!  ih0=oh/dh
  
  frequency: do iw=1,nw

     call sep_read(u0)
     u1=0.

     write(0,*) 'frequency',iw,'/',nw
     w = w0+(iw-1)*dw
     weight=(1-cmplx(0,1))*sqrt(w/(4*pi))

     shot: do is=1,ns
        offset: do ih=1,nh
!          do ihp=max(1,max(1-is-ih0,ih-nh-ih0)),min(nh,min(ih-1-ih0,ns-is-ih0))
           do ihp=max(1,max(1-is-ih0,ih-nh-ih0)),min(nh,min(ih-1+ih0,ns-is+ih0))
!              u1(ih,is)=u1(ih,is) + u0(ih-(ihp+ih0),is+(ihp+ih0))*u0(ihp,is)*weight
              u1(ih,is)=u1(ih,is) + u0(ih-(ihp-ih0),is+(ihp-ih0))*u0(ihp,is)*weight
           end do 
        end do offset
     end do shot

     call sep_write(u1)
  end do frequency
end program noah_conv












